package edu.stanford.nlp.optimization;

import edu.stanford.nlp.math.ArrayMath;

import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;

/**
 * <p>
 * Stochastic Gradient Descent Minimizer
 *
 *
 * The basic way to use the minimizer is with a null constructor, then
 * the simple minimize method:
 * <p/>
 * <p><code>Minimizer smd = new SGDMinimizer();</code>
 * <br><code>DiffFunction df = new SomeDiffFunction(); //Note that it must be a incidence of AbstractStochasticCachingDiffFunction</code>
 * <br><code>double tol = 1e-4;</code>
 * <br><code>double[] initial = getInitialGuess();</code>
 * <br><code>int maxIterations = someSafeNumber;
 * <br><code>double[] minimum = qnm.minimize(df,tol,initial,maxIterations);</code>
 * <p/>
 * Constructing with a null constructor will use the default values of
 * <p>
 * <br><code>batchSize = 15;</code>
 * <br><code>initialGain = 0.1;</code>
 * <p/>
 * <p/>
 *
 * @author <a href="mailto:akleeman@stanford.edu">Alex Kleeman</a>
 * @version 1.0
 * @since 1.0
 */
public class SGDMinimizer implements Minimizer {

  private int k;
  private int bSize = 15;
  private boolean quiet = false;

  public boolean outputIterationsToFile = true;
  public int outputFrequency = 10;
  public double gain = 0.1;
  private List<double[]> gradList = null;
  private int memory = 5;

  public void shutUp() {
    this.quiet = true;
  }

  public void setBatchSize(int batchSize) {
    bSize = batchSize;
  }


  private static NumberFormat nf = new DecimalFormat("0.000E0");

  public SGDMinimizer() {
  }

  public SGDMinimizer(double SGDGain, int batchSize){
    this.bSize = batchSize;
    this.gain = SGDGain;
  }


  /*
     This is the scaling factor for the gains to ensure convergence
   */
  private double gainSchedule(int it, double tau){
    return (tau / (tau + (double) it));
  }



  private double[] smooth(List<double[]> toSmooth){
    double[] smoothed = new double[toSmooth.get(0).length];

    for(double[] thisArray:toSmooth){
      ArrayMath.pairwiseAddInPlace(smoothed,thisArray);
    }

    ArrayMath.multiplyInPlace(smoothed,1/((double) toSmooth.size()));
    return smoothed;
  }



  public double[] minimize(Function function, double functionTolerance, double[] initial) {
    return minimize(function, functionTolerance, initial, -1);
  }


  public double[] minimize(Function function, double functionTolerance, double[] initial, int maxIterations) {
    sayln("SGDMinimizer called on function of " + function.domainDimension() + " variables;");
    sayln("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
    sayln("                     WARNING");
    sayln("As of now this does not have any convergence critera");
    sayln("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");


    // check for derivatives
    if (!(function instanceof DiffFunction)) {
      throw new UnsupportedOperationException();
    }


    //NEED TO MAKE SURE THAT THE FUNCTION IS ACTUALLY A STOCHASTIC FUNCTION
    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function;

    dfunction.method = StochasticCalculateMethods.GradientOnly;


    PrintWriter file = null;
    PrintWriter infoFile = null;
    if(outputIterationsToFile){
      try{
        file = new PrintWriter(new FileOutputStream("SGD" + bSize + "_g" + (gain* 1000) + ".output"),true);
        infoFile = new PrintWriter(new FileOutputStream("SGD" + bSize + "_g" + (gain* 1000) + ".info"),true);
      }
      catch (IOException e){
        System.err.println("Caught IOException outputing SGD data to file: " + e.getMessage());
        System.exit(1);
      }
    };


    double[] x, newX, grad, newGrad;
    int numBatches;
    long curTime;
    long calcTime;

    // initialize vector
    x = initial;

    grad = new double[x.length];
    newGrad = new double[x.length];
    newX = new double[x.length];
    gradList = new ArrayList<double[]>();


    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //            Initialization
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    numBatches = dfunction.dataDimension()/bSize;

    boolean have_max = (maxIterations > 0);

    if (maxIterations <= 0){
      maxIterations = 100*numBatches;
    }

    sayln("       Batchsize of: " + bSize);
    sayln("       Batches per pass through data:  " + numBatches );
    sayln("       Max iterations is = " + maxIterations);

    infoFile.println(function.domainDimension() + "; DomainDimension " );
    infoFile.println(bSize + "; batchSize ");
    infoFile.println(maxIterations + "; maxIterations");
    infoFile.println(numBatches + "; numBatches ");
    infoFile.println(dfunction.method.toString() + "; calculationMethod");
    infoFile.println(outputFrequency  + "; outputFrequency");
    //    double initGradNorm = ArrayMath.norm(grad);



    sayln("Iter: n ! Norm( HdotV ) ! Value ( aveImprovement)\n");

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //            Loop
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    curTime = System.currentTimeMillis();
    for (k = 0; k<maxIterations ; k++)  {

      double thisGain = gain*gainSchedule(k,5*numBatches);

      say("Iter: " + k + "  ++ " + (System.currentTimeMillis() - curTime) + " ms++  " );
      curTime = System.currentTimeMillis();


      //We know the values are new, so skip the equals check.

      if(k > 0 && gradList.size() >= memory){
        newGrad = gradList.remove(0);
      }else{
        newGrad = new double[grad.length];
      }

      dfunction.hasNewVals = true;
      System.arraycopy(dfunction.derivativeAt(x,bSize),0,newGrad,0,newGrad.length);

      gradList.add(newGrad);

      grad = smooth(gradList);

      //Get the next X
      for(int i = 0; i < x.length; i++){
        newX[i] = x[i] - thisGain*grad[i];
      }


      //if ((k > numBatches && size >= maxSize && averageImprovement / newValue < functionTolerance) ||
      if (k >= maxIterations) {
        sayln("    Tried to Stop");
        sayln("    Tolerence was  " + functionTolerance);
        x = newX;
        break;
      }

      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // THIS IS FOR DEBUG ONLY
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      if(outputIterationsToFile && (k%outputFrequency == 0)) {
        calcTime = (System.currentTimeMillis()- curTime);
        double curVal = dfunction.valueAt(x);

        say(" TrueValue{ " + curVal + " } ");

        file.println(k + " , " + curVal + " , " + thisGain + " , " + calcTime + " , " + ArrayMath.norm(ArrayMath.pairwiseSubtract(x,newX)));
      }

      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // END OF DEBUG STUFF
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      x = newX;


      if (quiet) {
        System.err.print(".");
      }else{
        sayln("");
      }


    }


    if(outputIterationsToFile){
      infoFile.println(" Completed after " + k + " Iterations");
      infoFile.println(" Final value of " + dfunction.valueAt(x));

      infoFile.close();
      file.close();
      System.err.println("Output Files Closed");
      //System.exit(1);

    }
    System.err.println("");
    System.err.println("Final Value is " + dfunction.valueAt(x));
    System.err.println("");
    return x;
  }


  private void sayln(String s) {
    if (!quiet) {
      System.err.println(s);
    }
  }

  private void say(String s) {
    if (!quiet) {
      System.err.print(s);
    }
  }




  public static void main(String[] args) {
    // optimizes test function using doubles and floats
    // test function is (0.5 sum(x_i^2 * var_i)) ^ PI
    // where var is a vector of random nonnegative numbers
    // dimensionality is variable.
    final int dim = 500000;
    final double maxVar = 5;
    final double[] var = new double[dim];
    double[] init = new double[dim];

    for (int i = 0; i < dim; i++) {
      init[i] = ((i + 1) / (double) dim - 0.5);//init[i] = (Math.random() - 0.5);
      var[i] = maxVar * (i + 1) / (double) dim;
    }

    final double[] grads = new double[dim];

    final DiffFunction f = new DiffFunction() {
      public double[] derivativeAt(double[] x) {
        double val = Math.PI * valuePow(x, Math.PI - 1);
        for (int i = 0; i < dim; i++) {
          grads[i] = x[i] * var[i] * val;
        }
        return grads;
      }

      public double valueAt(double[] x) {
        return 1.0 + valuePow(x, Math.PI);
      }

      private double valuePow(double[] x, double pow) {
        double val = 0.0;
        for (int i = 0; i < dim; i++) {
          val += x[i] * x[i] * var[i];
        }
        return Math.pow(val * 0.5, pow);
      }

      public int domainDimension() {
        return dim;
      }
    };



    SGDMinimizer min = new SGDMinimizer();

    min.minimize(f, 1.0E-4, init);

  }

}
