package edu.stanford.nlp.optimization;

import edu.stanford.nlp.math.ArrayMath;
import edu.stanford.nlp.sequences.SeqClassifierFlags;

import java.util.List;
import java.util.ArrayList;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;

/**
 * <p>
 * Stochastic Gradient Descent To Quasi Newton Minimizer
 *
 *  An experimental minimizer which takes a stochastic function (one implementing AbstractStochasticCachingDiffFunction)
 * and executes SGD for the first couple passes,  During the final iterations a series of approximate hessian vector
 * products are built up...  These are then passed to the QNminimizer so that it can start right up without the typical
 * delay.
 *
 * @author <a href="mailto:akleeman@stanford.edu">Alex Kleeman</a>
 * @version 1.0
 * @since 1.0
 */
public class SGDToQNMinimizer implements Minimizer<DiffFunction>, Serializable {

  private int k;
  private int bSize = 15;
  private boolean quiet = false;

  public boolean outputIterationsToFile = false;
  public int outputFrequency = 10;
  public double gain = 0.1;
  private List<double[]> gradList = null;
  private List<double[]> yList = null;
  private List<double[]> sList = null;
  private List<double[]> tmpYList = null;
  private List<double[]> tmpSList = null;
  private int memory = 5;
  public int SGDPasses = -1;
  public int QNPasses = -1;
  private int hessSampleSize = 50;
  private int QNMem = 10;
  private boolean toTest = false;

  public void shutUp() {
    this.quiet = true;
  }

  public void setBatchSize(int batchSize) {
    bSize = batchSize;
  }


  private static NumberFormat nf = new DecimalFormat("0.000E0");


  public SGDToQNMinimizer(SeqClassifierFlags flags){
    this.bSize = flags.stochasticBatchSize;
    this.gain = flags.initialGain;
    this.SGDPasses = flags.SGDPasses;
    this.QNPasses = flags.QNPasses;
    this.QNMem = flags.QNsize;
    this.outputIterationsToFile = flags.outputIterationsToFile;
    this.toTest = flags.testObjFunction;
    this.hessSampleSize = flags.SGD2QNhessSamples; 
  }


  public SGDToQNMinimizer(double SGDGain, int batchSize, int sgdPasses, int qnPasses, int hessSamples, int QNMem){
    this.bSize = batchSize;
    this.gain = SGDGain;
    this.SGDPasses = sgdPasses;
    this.QNPasses = qnPasses;
    this.QNMem = QNMem;
    this.hessSampleSize = hessSamples;
  }

  public SGDToQNMinimizer(double SGDGain, int batchSize, int SGDPasses, int QNPasses){
    this(SGDGain,batchSize,SGDPasses,QNPasses,50,10);
  }

  public void setQNMem(int mem){
    QNMem = mem;
  }

  public void setHessSampleSize(int size){
    hessSampleSize = size;
  }

  /*
     This is the scaling factor for the gains to ensure convergence
   */
  private double gainSchedule(int it, double tau){
    return (tau / (tau + (double) it));
  }



  private double[] smooth(List<double[]> toSmooth){
    double[] smoothed = new double[toSmooth.get(0).length];

    for(double[] thisArray:toSmooth){
      ArrayMath.pairwiseAddInPlace(smoothed,thisArray);
    }

    ArrayMath.multiplyInPlace(smoothed,1/((double) toSmooth.size()));
    return smoothed;
  }

  /*
  public double[] minimize(Function function, double functionTolerance, double[] initial) {
    return minimize(function,functionTolerance,initial,-1);
  }

  public double[] minimize(Function function, double functionTolerance, double[] initial, int maxIterations) {

    // check for stochastic derivatives
    if (!(function instanceof AbstractStochasticCachingDiffFunction)) {
      throw new UnsupportedOperationException();
    }
    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function;

    return minimize(dfunction, functionTolerance,initial,maxIterations);
  }

*/

  public double[] minimize(DiffFunction function, double functionTolerance, double[] initial) {
    return minimize(function,functionTolerance,initial,-1);
  }

  public double[] minimize(DiffFunction function, double functionTolerance, double[] initial, int maxIterations) {
    sayln("SGDToQNMinimizer called on function of " + function.domainDimension() + " variables;");

    // check for stochastic derivatives
    if (!(function instanceof AbstractStochasticCachingDiffFunction)) {
      throw new UnsupportedOperationException();
    }
    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function;

    if( toTest){
      StochasticDiffFunctionTester tester = new StochasticDiffFunctionTester(dfunction);
      if(tester.testSumOfBatches(initial,1e-4)){
        System.err.println("Testing complete... exiting");
        System.exit(1);
      }else{
        System.err.println("Testing failed....exiting");
        System.exit(1);
      };
    }

    dfunction.method = StochasticCalculateMethods.GradientOnly;


    //Set up for outputing iteration information to file
    PrintWriter file = null;
    PrintWriter infoFile = null;
    if(outputIterationsToFile){
      try{
        file = new PrintWriter(new FileOutputStream("SGD2QN" + bSize + "_g" + (gain* 1000) + ".output"),true);
        infoFile = new PrintWriter(new FileOutputStream("SGD2QN" + bSize + "_g" + (gain* 1000) + ".info"),true);
      }
      catch (IOException e){
        System.err.println("Caught IOException outputing SGD data to file: " + e.getMessage());
        System.exit(1);
      }
    };


    // Declare variables needed during iteration
    double[] x, newX, grad, newGrad,newHx;
    int numBatches;
    long curTime;
    long calcTime;

    // initialize vector
    x = initial;

    grad = new double[x.length];
    newGrad = new double[x.length];
    newX = new double[x.length];
    newHx = new double[x.length];

    gradList = new ArrayList<double[]>();
    yList = new ArrayList<double[]>();
    sList = new ArrayList<double[]>();
    tmpYList = new ArrayList<double[]>();
    tmpSList = new ArrayList<double[]>();

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //            Initialization
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    if ( SGDPasses < 0 || QNPasses < 0){
      System.err.println("Error:  Number of iterations not set");
      System.exit(1);
    }

    numBatches = dfunction.dataDimension()/bSize;
    int itsBeforeSecondOrderInfo = numBatches*SGDPasses - hessSampleSize * QNMem;

    sayln("       Batchsize of: " + bSize);
    sayln("       Batches per pass through data:  " + numBatches );
    sayln("       Passes of SGD: "+ SGDPasses);
    sayln("       Passes of QN:  " + QNPasses);
    sayln("       Hess sample size:   " + hessSampleSize);
    sayln("       QNSize:   " + QNMem);
    sayln("       ");
    if (itsBeforeSecondOrderInfo < 0){
      System.err.println("Not enough data for mem and smoothing settings");
      System.exit(1);
    }



    int SGDIterations = numBatches*SGDPasses;

    if(outputIterationsToFile){
      infoFile.println(dfunction.domainDimension() + "; DomainDimension " );
      infoFile.println(bSize + "; batchSize ");
      infoFile.println(SGDPasses + "; SGDPasses");
      infoFile.println(QNPasses + "; QNPasses");
      infoFile.println(numBatches + "; numBatches ");
      infoFile.println(hessSampleSize + ";hessSampleSize");
      infoFile.println(QNMem + ";QNSize");
      //infoFile.println(dfunction.method.toString() + "; calculationMethod");
      infoFile.println(outputFrequency  + "; outputFrequency");
    }
    //    double initGradNorm = ArrayMath.norm(grad);



    sayln("Iter: n ++TimeforLastIt++\n");

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //            Loop
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    curTime = System.currentTimeMillis();
    for (k = 0; k<SGDIterations ; k++)  {

      double thisGain = gain*gainSchedule(k,5*numBatches);

      say("Iter: " + k + "  ++ " + (System.currentTimeMillis() - curTime) + " ms++  " );
      curTime = System.currentTimeMillis();


      //We know the values are new, so skip the equals check.

      if(k > 0 && gradList.size() >= memory){
        newGrad = gradList.remove(0);
      }else{
        newGrad = new double[grad.length];
      }

      if(k > itsBeforeSecondOrderInfo){
        newHx = new double[grad.length];

        dfunction.method = StochasticCalculateMethods.IncorporatedFiniteDifference;
        //dfunction.method = StochasticCalculateMethods.ExternalFiniteDifference;
        dfunction.hasNewVals = true;
        System.arraycopy(dfunction.derivativeAt(x,x,bSize),0,newGrad,0,newGrad.length);
        dfunction.returnPreviousValues = true;
        System.arraycopy(dfunction.HdotVAt(x,x,bSize),0,newHx,0,grad.length);

        tmpYList.add(newHx);
        tmpSList.add(newX);

        if(tmpYList.size() == hessSampleSize){
          yList.add(smooth(tmpYList));
          sList.add(smooth(tmpSList));
          tmpYList.clear();
          tmpSList.clear();
        }

      } else{
        dfunction.hasNewVals = true;
        System.arraycopy(dfunction.derivativeAt(x,bSize),0,newGrad,0,newGrad.length);
      }

      gradList.add(newGrad);

      grad = smooth(gradList);

      //Get the next X
      for(int i = 0; i < x.length; i++){
        newX[i] = x[i] - thisGain*grad[i];
      }

      //If we've run through the data SGDPasses number of times then continue to QN
      if (k >= SGDIterations) {
        sayln("SGD completed.");
        x = newX;
        break;
      }

      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // THIS IS FOR DEBUG ONLY
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      if(outputIterationsToFile && (k%outputFrequency == 0)) {
        calcTime = (System.currentTimeMillis()- curTime);
        double curVal = dfunction.valueAt(x);
        say(" TrueValue{ " + curVal + " } ");
        file.println(k + " , " + curVal + " , " + thisGain + " , " + calcTime + " , " + ArrayMath.norm(ArrayMath.pairwiseSubtract(x,newX)));
      }
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // END OF DEBUG STUFF
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      x = newX;

      if (quiet) {
        System.err.print(".");
      }else{
        sayln("");
      }

    }

    System.err.println("Passing off to QN");
    System.err.println("");
    QNMinimizer qnMin = new QNMinimizer(QNMem);
    qnMin.setHistory(sList,yList);
    x = qnMin.minimize(dfunction,functionTolerance,x,QNPasses);

    if(outputIterationsToFile){
      infoFile.close();
      file.close();
      System.err.println("Output Files Closed");
    }

    System.err.println("");
    System.err.println("QN Minimization complete.");
    System.err.println("");
    return x;
  }


  private void sayln(String s) {
    if (!quiet) {
      System.err.println(s);
    }
  }

  private void say(String s) {
    if (!quiet) {
      System.err.print(s);
    }
  }





}
