package edu.stanford.nlp.optimization;

import edu.stanford.nlp.math.ArrayMath;

import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.util.Random;
import java.util.Arrays;
import java.io.*;

/**
 * <p>
 * Stochastic Meta Descent Minimizer based on
 *
 * <p>
 * Accelerated training of conditional random fields with stochastic gradient methods
 * S. V. N. Vishwanathan, Nicol N. Schraudolph, Mark W. Schmidt, Kevin P. Murphy
 * June 2006 	 	Proceedings of the 23rd international conference on Machine learning ICML '06
 * Publisher: ACM Press
 * <p/>
 * The basic way to use the minimizer is with a null constructor, then
 * the simple minimize method:
 * <p/>
 * <p><code>Minimizer smd = new SMDMinimizer();</code>
 * <br><code>DiffFunction df = new SomeDiffFunction();</code>
 * <br><code>double tol = 1e-4;</code>
 * <br><code>double[] initial = getInitialGuess();</code>
 * <br><code>int maxIterations = someSafeNumber;
 * <br><code>double[] minimum = qnm.minimize(df,tol,initial,maxIterations);</code>
 * <p/>
 * Constructing with a null constructor will use the default values of
 * <p>
 * <br><code>batchSize = 15;</code>
 * <br><code>initialGain = 0.1;</code>
 * <br><code>useAlgorithmicDifferentiation = true;</code>
 * <p/>
 * <p/>
 *
 * @author <a href="mailto:akleeman@stanford.edu">Alex Kleeman</a>
 * @version 1.0
 * @since 1.0
 */
public class SMDMinimizer implements Minimizer {

  private int k;
  private int bSize = 15;
  private boolean quiet = false;
  private StochasticCalculateMethods methodSMD = StochasticCalculateMethods.NoneSpecified;

  public boolean outputIterationsToFile = true;
  public int outputFrequency = 10;
  public double initialGain = 0.1;
  public boolean restrictSteps = false;
  public boolean useAlgorithmicDifferentiation = true;
  PrintWriter file = null;
  PrintWriter infoFile = null;
  public double mu = 0.1;
  public double lam = 1;
  public double cPosDef = 0.01;
  //DEBUG ONLY
  public boolean testObjFunc = false;
  public boolean printMinMax = false;
  public double finalVal = 0.0;
  public boolean useGaussNewton = false;


  public void shutUp() {
    this.quiet = true;
  }

  public void setBatchSize(int batchSize) {
    bSize = batchSize;
  }


  private static NumberFormat nf = new DecimalFormat("0.000E0");

  public SMDMinimizer() {
  }

  public SMDMinimizer(double initialSMDGain, int batchSize,StochasticCalculateMethods method){
    this.bSize = batchSize;
    this.initialGain = initialSMDGain;
    this.methodSMD = method;
  }




  public double[] minimize(Function function, double functionTolerance, double[] initial) {
    return minimize(function, functionTolerance, initial, -1);
  }


  public double[] minimize(Function function, double functionTolerance, double[] initial, int maxIterations) {
    sayln("SMDMinimizer called on function of " + function.domainDimension() + " variables;");

    /*Random considerations      -akleeman

    What is the best way to gauge convergence?  Calculating the full value, or the full gradient will
      clearly not a be a good idea (at least not every iteration) since that will take considerably longer
      than the stochastic gradient calculation.

    The only real thing we have that gives some sense of convergence is the position vector x[].  But
      getting any information (norm(newX -X) etc..) out of this will again add to computation time considerably.

    Can some sort of estimate of the importance of each batch be made based on the value calculated off of that
      set of data, in which case on the second run through the data you can sort and focus on the important batches?
      (i.e.) can we assume that a large value calculation on a set of data means that it has a large influence on the
      objective function value?

    Is it worth it to do AD?  Or is there a more efficient way of writting the AD classes that will reduce the
      overhead?  As it stands the AD calcuation in the LogConditionalLikelihood objective function requires two loops
      of length of x in size.  It seems that this is where most of the compuation time ends up.  Can we eliminate
      these loops some how?  (caching, reusing old batch information etc..)

    Parelelization??  It seems like this would be extremely easy to turn into a parallel operation.
      at the very least you could parellelize the update loop, as well as the objective function loops I talked about
      above.



    more questions:

    - Why do things go haywire when batchSizes over about 50 are used?  Running with batch size of 30 seems to work
    */


    // check for derivatives
    if (!(function instanceof DiffFunction)) {
      throw new UnsupportedOperationException();
    }

    //NEED TO MAKE SURE THAT THE FUNCTION IS ACTUALLY A STOCHASTIC FUNCTION
    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function;
  

    /*if(useGaussNewton){
      dfunction.method = StochasticCalculateMethods.GradientOnly;
    }*/

    // for convergence test
    //Queue<Double> previousVals = new LinkedList<Double>();
    if (methodSMD.equals(StochasticCalculateMethods.NoneSpecified)){
      System.err.println("No method has been set for the Stochastic Calculation");
      System.exit(1);
    }

    dfunction.method = methodSMD;


    /*  THIS RUNS A FULL SET OF TESTS ON THE OBJECTIVE FUNCITON
        -Checks to see if sum of batches = full
        -Checks different methods of Hessian Vector product and compares outcomes.
    */

    if (testObjFunc)
    {
      ArrayMath.addInPlace(initial,0.1);
      testObjectiveFunction(function,initial,functionTolerance);
      System.exit(1);

    }


    double[] x, newX, grad, newGrad,v,gain,Hv,G;
    double meta;
    int numBatches;
    long curTime;
    long calcTime;


    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //            Initialization
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    x = initial;

    gain = new double[x.length];
    v = new double[x.length];
    grad = new double[x.length];
    newGrad = new double[x.length];
    newX = new double[x.length];
    Hv = new double[x.length];
    G = new double[x.length];


    //Initialize arrays
    System.arraycopy(dfunction.derivativeAt(x,v,bSize ), 0,grad, 0, grad.length);

    for(int i = 0; i<v.length;i++){
      gain[i] = initialGain;
      x[i] -= initialGain*grad[i];
      v[i] = -initialGain*grad[i];
    }

    curTime = System.currentTimeMillis();


    numBatches = dfunction.dataDimension()/bSize;

    boolean have_max = (maxIterations > 0);

    if (maxIterations <= 0){
      maxIterations = 100*numBatches;
    }

    sayln("       Batchsize of: " + bSize);
    sayln("       Batches per pass through data:  " + numBatches );
    sayln("       Max iterations is = " + maxIterations);
    sayln("       All calculations will be made using " + dfunction.method.toString() );


    if (outputIterationsToFile){

      outputInitializeFiles();

      infoFile.println(function.domainDimension() + "; DomainDimension " );
      infoFile.println(bSize + "; batchSize ");
      infoFile.println(cPosDef + "; c");
      infoFile.println(lam + "; lambda ");
      infoFile.println(initialGain + "; initGain");
      infoFile.println(maxIterations + "; maxIterations");
      infoFile.println(numBatches + "; numBatches ");
      infoFile.println(dfunction.method.toString() + "; calculationMethod");
      infoFile.println(outputFrequency  + "; outputFrequency");
    }
    //    double initGradNorm = ArrayMath.norm(grad);



    sayln("Iter: n [time ms] Value \n");



    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //            Start of Loop
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    // Get the current gradient and H.v
    System.arraycopy(dfunction.derivativeAt(x,v,bSize ), 0,grad, 0, grad.length);
    dfunction.returnPreviousValues = true;
    System.arraycopy(dfunction.HdotVAt(x,v,grad,bSize), 0, Hv, 0, Hv.length);
    calcTime = System.currentTimeMillis();

    for (k = 0; k<maxIterations ; k++)  {

      double newValue = 0;

      say("Iter: " + k + " [ " + (System.currentTimeMillis() - curTime) + " ms]  " );
      curTime = System.currentTimeMillis();



      //Update the weights
      for(int i = 0; i < x.length; i++){
        meta = 1-mu*grad[i]*v[i];

        if(0.5 > meta){
          gain[i] = gain[i]*0.5;
        }else{
          gain[i] = gain[i]*meta;
        }

        // NOW BEGINING iteration i+1
        //----------------------------
        //Update gain history
        v[i] = lam*(1+cPosDef*gain[i])*v[i] - gain[i]*(grad[i] + lam*Hv[i]);
        //Get the next X
        newX[i] = x[i] - gain[i]*grad[i];
      }

      if(printMinMax){
        say("vMin = " + ArrayMath.min(v) + "  ");
        say("vMax = " + ArrayMath.max(v) + "  ");
        say("gainMin = " + ArrayMath.min(gain) + "  ");
        say("gainMax = " + ArrayMath.max(gain) + "  ");
      }


      //We know the values are new, so skip the comparison.
      dfunction.hasNewVals = true;

      System.arraycopy(dfunction.derivativeAt(newX,v,bSize),0,newGrad,0,newGrad.length);
      newValue = dfunction.value;

      dfunction.returnPreviousValues = true;  //Same deal Don't even try and compare x and v we know they're the same
      System.arraycopy(dfunction.HdotVAt(newX,v,newGrad,bSize) ,0,Hv,0,Hv.length);


      
      if(useGaussNewton){
        //System.err.print("used GN Approx");
        G = ArrayMath.multiply(newGrad, ArrayMath.innerProduct(newGrad,v) );

        say(nf.format( ArrayMath.norm(ArrayMath.pairwiseSubtract(Hv,G))) + "  G-H " );

        say(nf.format(newValue));

        if(newValue < 5){
          Hv = G;
        } else{
          System.err.print(" switched ");
        }
      }


      if ((have_max && k >= maxIterations)) {
        sayln("    Tried to Stop");
        x = newX;
        break;
      }

      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // THIS IS FOR DEBUG ONLY   -if you want to skip this step, just set outputIterationsToFile=false in the header.
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


      if(outputIterationsToFile && (k%outputFrequency == 0)) {
        calcTime = (System.currentTimeMillis()- calcTime);
        double curVal = dfunction.valueAt(x);

        say(" TrueValue{ " + curVal + " } ");
        file.println(k + " , " + curVal + " , " + newValue + " , " + calcTime + " , " + ArrayMath.norm(ArrayMath.pairwiseSubtract(x,newX)));
        //file.println(" , " + ArrayMath.min(x) + " , " + ArrayMath.max(x) + " , " + ArrayMath.min(gain) + " , " + ArrayMath.max(gain) + " , " + ArrayMath.min(v) + " , " + ArrayMath.max(v) + " , ");
        calcTime = System.currentTimeMillis();
      }

      sayln("");

      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // END OF DEBUG STUFF
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



      System.arraycopy(newGrad, 0, grad, 0, newGrad.length);
      x = newX;

      if (quiet) {
        System.err.print(".");
      }


    }


    if(outputIterationsToFile){
      finalVal = dfunction.valueAt(x);
      infoFile.println(" iterationCount ; " + k);
      infoFile.println(" finalValue ; " + finalVal);
      infoFile.println(finalVal);
      infoFile.close();
      file.close();
      System.err.println("Output Files Closed");

    }

    return x;
  }




  private void outputInitializeFiles(){

    if(outputIterationsToFile){
      try{
        file = new PrintWriter(new FileOutputStream("SMD_" + methodSMD.toString() + "_mu" + mu*1000 + "_lam" + lam*1000 + "_b" + bSize + "_g" + (initialGain* 1000) + ".output"),true);
        infoFile = new PrintWriter(new FileOutputStream("SMD_" + methodSMD.toString() + "_mu" + mu*1000 + "_lam" + lam*1000 + "_b" + bSize + "_g" + (initialGain* 1000) + ".info"),true);
      }
      catch (IOException e){
        System.err.println("Caught IOException outputing SMD data to file: " + e.getMessage());
        System.exit(1);
      }
    };
  }




  private static class SurpriseConvergence extends Throwable {
    public SurpriseConvergence(String s) {
      super(s);
    }
  }

  private void sayln(String s) {
    if (!quiet) {
      System.err.println(s);
    }
  }

  private void say(String s) {
    if (!quiet) {
      System.err.print(s);
    }
  }





















  //  Get Prime Factors of an integer ....
  //  Code was originally from    http://www.idinews.com/sourcecode/IntegerFunction.html
  //  Decompose integer into prime factors
  //  ------------------------------------

  //  Upon return result[0] contains the number of factors (0 if N is 0), and
  //  result[1] . . . result[result[0]] contain the factors in ascending order.

  public static long[] primeFactors(long N)
  {long [] fctr = new long[64];       //  Result array
    long n = Math.abs(N);              //  Guard against negative

    short fctrIndex = 0;

    if (n > 0) {                       //  Guard against zero

      //  First do special cases 2 and 3

      while (n % 2 == 0)  {fctr[++fctrIndex] = 2; n /= 2;}
      while (n % 3 == 0)  {fctr[++fctrIndex] = 3; n /= 3;}

      //  Then every 6n-1 and 6n+1 until the divisor exceeds the square root
      //  of the current quotient.  NOTE:  Some trial divisors will be
      //  non-primes, e.g. 25, 35, 49, 55.  They have no effect, however,
      //  since their prime factors will already have been tried.

      for (int k = 5; k*k <= n; k += 6)
        for (int dvsr = k; dvsr <= k+2; dvsr+=2)
        { while (n % dvsr == 0)
        {fctr[++fctrIndex] = dvsr;  n /= dvsr;}
        }

      if (n > 1) fctr[++fctrIndex] = n; //  Store final factor, if any
    }

    fctr[0] = fctrIndex;                //  Store number of factors
    return fctr;
  }


  /**
   *      getTestBatchSize - This function takes as input the size of the data and returns the largest factor of the data size
   *    this is done so that when testing the function we are gaurenteed to have equally sized batches, and that the fewest
   *    number of evaluations needs to be made in order to test the function.
   *
   * @param size   - The size of the current data set
   * @return         The largest factor of the data size
   */
  private static long getTestBatchSize(long size){

    long testBatchSize = 1;

    long[] factors = primeFactors( size );

    long factorCount = factors[0];

    // Calculate the batchsize for the factors
    if( factorCount == 0 ){
      System.err.println("Attempt to test function on data of prime dimension.  This would involve a batchSize of 1 and may take a very long time.");
      System.exit(1);
    }else if (factorCount == 2){
      testBatchSize = (int) factors[1];
    }else {
      //  find the largest factor.
      for( int f = 1; f< factorCount;f++){
        testBatchSize *= factors[f];
      }
    }

    return testBatchSize;
  }

  /**
   * testObjectiveFunction
   *  This function was written to provide a test for accuracy of stochastic objective functions.  The test
   *  checks for the following properties:
   *
   * 1) The sum of the value over each batch equals the full value
   * 2) The sum of the gradients over each batch equals the full gradient
   * 3) The gradient calculated using Incorporated Finite Difference is never more than functionTolerance from the
   *        gradient using External Finite Difference
   * 4) The hessian vector also does not varry between Incorporated and External Finite Difference
   * 
   * @param function          The function to test
   * @param x                 The point to use for testing (v is generated randomly
   * @param functionTolerance The tolerance
   * @return
   */

  public boolean testObjectiveFunction(Function function, double[] x, double functionTolerance){

    double[]  approxGrad,fullGrad,diff,HvAD,HvFD,v,curGrad,gradFD;
    double diffNorm,diffValue,fullValue,approxValue,diffGrad,maxGradDiff = 0.0,maxHvDiff = 0.0;


    Random generator = new Random(System.currentTimeMillis());  // used to generate random test vectors
    v = new double[x.length];
    HvAD = new double[x.length];
    HvFD = new double[x.length];
    gradFD = new double[x.length];
    fullGrad = new double[x.length];

    // check for derivatives
    if (!(function instanceof DiffFunction)) {
      throw new UnsupportedOperationException();
    }


    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function; // Make sure the function is Stochastic

    //  Look for a good batchSize to test with by getting factors
    int testBatchSize = (int) getTestBatchSize(dfunction.dataDimension());

    //  Again make sure that our calculated batchSize is actually valid
    if(testBatchSize < 0 || testBatchSize > dfunction.dataDimension() || (dfunction.dataDimension()%testBatchSize != 0)){
      System.err.println("Invalid testBatchSize found, testing aborted.  Data size: " + dfunction.dataDimension() + " batchSize: " + testBatchSize);
      System.exit(1);
    }

    int numBatches = dfunction.dataDimension()/testBatchSize;

    sayln("Testing Gradients");
    sayln("data dimension  = " + dfunction.dataDimension());
    sayln("batch size = " + testBatchSize);
    sayln("number of batches = " + numBatches);


    approxGrad = new double[x.length];
    curGrad = new double[x.length];
    approxValue = 0;

    //Generate the initial vectors
    for (int i = 0; i < x.length; i ++){
      approxGrad[i] = 0;
      v[i] = generator.nextDouble() ;
    }


    //This loop runs through all the batches and sums of the calculations to compare against the full gradient
    for (int i = 0; i < numBatches ; i ++){

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Perform calculation using IncorporatedFiniteDifference
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      
      dfunction.method = StochasticCalculateMethods.IncorporatedFiniteDifference;

      //  update the value
      approxValue += dfunction.valueAt(x,v,testBatchSize);

      //  update the gradient
      dfunction.returnPreviousValues = true;
      System.arraycopy(dfunction.derivativeAt(x,v,testBatchSize ), 0,curGrad, 0, curGrad.length);

      //  update the Hessian
      dfunction.returnPreviousValues = true;
      System.arraycopy(dfunction.HdotVAt(x,v,testBatchSize),0,HvAD,0,HvAD.length);

      //Update Approximate
      approxGrad = ArrayMath.pairwiseAdd(approxGrad,curGrad);

 // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 // Perform calculations using external finite difference
 // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      dfunction.method = StochasticCalculateMethods.ExternalFiniteDifference;

      dfunction.recalculatePrevBatch = true;
      System.arraycopy(dfunction.derivativeAt(x,v,testBatchSize ), 0,gradFD, 0, gradFD.length);

      dfunction.recalculatePrevBatch = true;
      System.arraycopy(dfunction.HdotVAt(x,v,gradFD,testBatchSize),0,HvFD,0,HvFD.length);

      double DiffGrad = ArrayMath.norm_inf(ArrayMath.pairwiseSubtract(gradFD,curGrad));

      // Keep track of the biggest error.
      if (DiffGrad > maxGradDiff){maxGradDiff = DiffGrad;}

      double DiffHv = ArrayMath.norm_inf(ArrayMath.pairwiseSubtract(HvAD,HvFD));

      //Keep track of the biggest H.v error
      if (DiffHv > maxHvDiff){maxHvDiff = DiffHv;}
    }

 // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 // Get the full gradient and value, these should equal the approximates
 // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    System.arraycopy(dfunction.derivativeAt(x),0,fullGrad,0,fullGrad.length);
    fullValue = dfunction.valueAt(x);


    if(ArrayMath.norm_inf(ArrayMath.pairwiseSubtract(fullGrad,approxGrad)) < functionTolerance){
      sayln("");
      sayln("  Gradient is looking good");
    }else{
      diff = new double[x.length];
      diff = ArrayMath.pairwiseSubtract(approxGrad,fullGrad);
      diffNorm = ArrayMath.norm(diff);
      sayln("");
      sayln("  Seems there is a problem.  Gradient is off by norm of " + diffNorm);
    };

    if( maxGradDiff < functionTolerance ){
      sayln("");
      sayln("  Both gradients are the same");
    }else{
      diffValue = approxValue - fullValue;
      sayln("");
      sayln("  Seems there is a problem.  The two methods of calculating the gradient are different  max |AD-FD|_inf Error of " + maxGradDiff);
    };


    if( Math.abs(fullValue - approxValue) < functionTolerance){
      sayln("");
      sayln("  Value is looking good");
    }else{
      diffValue = approxValue - fullValue;
      sayln("");
      sayln("  Seems there is a problem.  Value is off by " + diffValue);
    };

    if(maxHvDiff < functionTolerance){
      sayln("");
      sayln("  Hv Approimations line up well");
    }else{
      sayln("");
      sayln("    Seems there is a problem.  Hv approximations aren't quite close enough -- max |AD-FD|_inf Error of " + maxHvDiff);
    }

    return true;
  }





  /*
  public static void outputFullHess(Function function, double[] x)  {

    double[]  v,Hv;


    // check for derivatives
    if (!(function instanceof DiffFunction)) {
      throw new UnsupportedOperationException();
    }
    //DiffFunction dfunction = (DiffFunction) function;
    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function;



    Hv = new double[x.length];
    v = new double[x.length];

    //Initialize arrays
    for(int i = 0; i<x.length;i++){
      v[i] = 0;
    }

    try{
      PrintWriter file = new PrintWriter(new FileOutputStream("test.tmp"),true);

      int itSize = 1000;

      for(int i = 0; i < itSize;i++){
        if(i>0)  v[i-1]=0;

        v[i] = 1;
        System.arraycopy(dfunction.HdotVAt(x,v,dfunction.dataDimension() ), 0, Hv, 0, Hv.length);

        for(int j = 0; j <itSize;j++){
          file.print(" " + Hv[j] + " ");
        }
        file.print("\n");


      }
      file.close();
      System.exit(1);
    }
    catch (IOException e){
      System.err.println("Caught IOException in outputFullHess: " + e.getMessage());
      System.exit(1);
    }



  }
*/




  public static void main(String[] args) {
    // optimizes test function using doubles and floats
    // test function is (0.5 sum(x_i^2 * var_i)) ^ PI
    // where var is a vector of random nonnegative numbers
    // dimensionality is variable.
    final int dim = 500000;
    final double maxVar = 5;
    final double[] var = new double[dim];
    double[] init = new double[dim];

    for (int i = 0; i < dim; i++) {
      init[i] = ((i + 1) / (double) dim - 0.5);//init[i] = (Math.random() - 0.5);
      var[i] = maxVar * (i + 1) / (double) dim;
    }

    final double[] grads = new double[dim];

    final DiffFunction f = new DiffFunction() {
      public double[] derivativeAt(double[] x) {
        double val = Math.PI * valuePow(x, Math.PI - 1);
        for (int i = 0; i < dim; i++) {
          grads[i] = x[i] * var[i] * val;
        }
        return grads;
      }

      public double valueAt(double[] x) {
        return 1.0 + valuePow(x, Math.PI);
      }

      private double valuePow(double[] x, double pow) {
        double val = 0.0;
        for (int i = 0; i < dim; i++) {
          val += x[i] * x[i] * var[i];
        }
        return Math.pow(val * 0.5, pow);
      }

      public int domainDimension() {
        return dim;
      }
    };



    SMDMinimizer min = new SMDMinimizer();

    min.minimize(f, 1.0E-4, init);

  }

}


