package edu.stanford.nlp.optimization;

import edu.stanford.nlp.math.ArrayMath;
import edu.stanford.nlp.linalg.Array;

import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.util.List;
import java.util.ArrayList;
import java.util.Queue;
import java.util.LinkedList;

/**
 * Online Limited-Memory Quasi-Newton BFGS implementation based on the algorithms in
 * <p>
 * Nocedal, Jorge, and Stephen J. Wright.  2000.  Numerical Optimization.  Springer.  pp. 224--
 * <p>
 * and modified to the online version presented in
 * <p>
 * A Stocahstic Quasi-Newton Method for Online Convex Optimization
 * Schraudolph, Yu, Gunter (2007)
 * <p>
 * As of now, it requires a
 * Stochastic differentiable function (AbstractStochasticCachingDiffFunction) as input.
 * <p/>
 * The basic way to use the minimizer is with a null constructor, then
 * the simple minimize method:
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * THIS IS NOT UPDATE FOR THE STOCHASTIC VERSION YET.
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * <p/>
 * <p><code>Minimizer qnm = new QNMinimizer();</code>
 * <br><code>DiffFunction df = new SomeDiffFunction();</code>
 * <br><code>double tol = 1e-4;</code>
 * <br><code>double[] initial = getInitialGuess();</code>
 * <br><code>double[] minimum = qnm.minimize(df,tol,initial);</code>
 * <p/>
 * <p/>
 * If you do not choose a value of M, it will use the max amount of memory
 * available, up to M of 20.  This will slow things down a bit at first due
 * to forced garbage collection, but is probably faster overall b/c you are
 * guaranteed the largest possible M.
 *
 * The Stochastic version was written by Alex Kleeman, but about 95% of the code
 * was taken directly from the previous QNMinimizer written mostly by Jenny.
 *
 * @author <a href="mailto:jrfinkel@stanford.edu">Jenny Finkel</a>
 * @author Galen Andrew
 * @author <a href="mailto:akleeman@stanford.edu">Alex Kleeman</a>
 * @version 1.0
 * @since 1.0
 */
public class SQNMinimizer implements Minimizer<DiffFunction> {

  private int k;
  private int M = 0;
  private double lambda = 0;
  private double initGain = 1;
  private double tau = 1e4;
  private int batchSize = 200;
  private double cPosDef = 1;
  private double epsilon = 1e-10;
  PrintWriter file = null;
  PrintWriter infoFile = null;

  public boolean outputIterationsToFile = true;
  public int outputFrequency = 5;
  public double calcTime;

  private boolean quiet = false;

  public void shutUp() {
    this.quiet = true;
  }

  public void setM(int m) {
    M = m;
  }

  private Function monitor;
  private FloatFunction floatMonitor;

  private static NumberFormat nf = new DecimalFormat("0.000E0");

  public SQNMinimizer(int m) {
    M = m;
  }

  public SQNMinimizer() {
  }

  public SQNMinimizer(int mem,double initialGain, int batchSize) {
    this.initGain = initialGain;
    this.batchSize = batchSize;
    this.M = mem;
  }

  public SQNMinimizer(Function monitor, int m) {
    this.monitor = monitor;
    M = m;
  }

  public SQNMinimizer(FloatFunction monitor) {
    this.floatMonitor = monitor;
  }

  // computes d = a + b * c
  private static double[] plusAndConstMult(double[] a, double[] b, double c, double[] d) {
    for (int i = 0; i < a.length; i++) {
      d[i] = a[i] + c * b[i];
    }
    return d;
  }


  /*
     This is the scaling factor for the gains to ensure convergence
   */
  private double gainSchedule(int it, double tau){
    return (tau / (tau + (double) it));
  }


  private List<double[]> sList = new ArrayList<double[]>();
  private List<double[]> yList = new ArrayList<double[]>();
  private List<Double> roList = new ArrayList<Double>();

  private void computeDir(double[] dir, double[] fg) throws SQNMinimizer.SurpriseConvergence {
    System.arraycopy(fg, 0, dir, 0, fg.length);

    int mmm = sList.size();
    double[] as = new double[mmm];
    double[] factors = new double[dir.length];

    for (int i = mmm - 1; i >= 0; i--) {
      as[i] = roList.get(i) * ArrayMath.innerProduct(sList.get(i), dir);
      plusAndConstMult(dir, yList.get(i), -as[i], dir);
    }

    // multiply by hessian approximation
    if (mmm != 0) {
      double[] y = yList.get(mmm - 1);
      double yDotY = ArrayMath.innerProduct(y, y);
      if (yDotY == 0) {
        throw new SQNMinimizer.SurpriseConvergence("Y is 0!!");
      }
      double gamma = ArrayMath.innerProduct(sList.get(mmm - 1), y) / yDotY;
      ArrayMath.multiplyInPlace(dir, gamma);
    }else if(mmm == 0){
      //This is a safety feature preventing too large of an initial step (see Yu Schraudolph Gunter)
      ArrayMath.multiplyInPlace(dir,epsilon);
    }

    for (int i = 0; i < mmm; i++) {
      double b = roList.get(i) * ArrayMath.innerProduct(yList.get(i), dir);
      plusAndConstMult(dir, sList.get(i), cPosDef*as[i] - b, dir);
      plusAndConstMult(ArrayMath.pairwiseMultiply(yList.get(i),sList.get(i)),factors,1,factors);
    }

    ArrayMath.multiplyInPlace(dir, -1);
  }

  public double[] minimize(DiffFunction function, double functionTolerance, double[] initial) {
			return minimize(function, functionTolerance, initial, -1);
	}

	public double[] minimize(DiffFunction function, double functionTolerance, double[] initial, int maxIterations) {
    say("Stochastic QNMinimizer called on double function of " + function.domainDimension() + " variables;");
    if (M > 0) {
      sayln(" Using m = " + M);
    } else {
      sayln(" Using dynamic setting of M.");
    }


    AbstractStochasticCachingDiffFunction dfunction = (AbstractStochasticCachingDiffFunction) function;

    //dfunction.method = StochasticCalculateMethods.GradientOnly;
    dfunction.method = StochasticCalculateMethods.IncorporatedFiniteDifference ;

    double numBatches = dfunction.dataDimension()/batchSize;

    sayln("       Batchsize of: " + batchSize);
    sayln("       Batches per pass through data:  " + numBatches );
    sayln("       Max iterations is = " + maxIterations);
    sayln("       All calculations will be made using " + dfunction.method.toString() );





    if (outputIterationsToFile){

      outputInitializeFiles();

      //infoFile.println("DomainDimension , batchSize, c , lambda , initGain, maxIterations, numBatches, calculationMethod, outputFrequency ");
      infoFile.println(function.domainDimension() + "; DomainDimension " );
      infoFile.println(batchSize + "; batchSize ");
      infoFile.println(cPosDef + "; c");
      infoFile.println(lambda + "; lambda ");
      infoFile.println(initGain + "; initGain");
      infoFile.println(maxIterations + "; maxIterations");
      infoFile.println(numBatches + "; numBatches ");
      infoFile.println(dfunction.method.toString() + "; calculationMethod");
      infoFile.println(outputFrequency  + "; outputFrequency");
    }



    // for convergence test
    Queue<Double> previousVals = new LinkedList<Double>();

    double[] x, newX, grad, newGrad,newGradSameBatch, dir, gainVect;
    double ro,gain;
    //int i;

    //For Gain Adapatation
    double[] Hdir,v;

    // initialize matrix
    x = initial;


    if (monitor != null) {
      System.err.println("Monitor not supported yet.");
      System.exit(1);
      //monitor.valueAt(x);
    }

    // initialize gradient
    grad = new double[x.length];
    gainVect = new double[x.length];
    v = new double[x.length];
    Hdir = new double[x.length];


    dfunction.returnPreviousValues = true;
    System.arraycopy(dfunction.derivativeAt(x,batchSize), 0, grad, 0, grad.length);
    newGrad = new double[x.length];
    newGradSameBatch = new double[x.length];

    for(int i = 0; i<v.length;i++){
      gainVect[i] = initGain;
      x[i] -= initGain*grad[i];
      v[i] = -initGain*grad[i];
    }

    newX = new double[x.length];
    dir = new double[x.length];

    sList = new ArrayList<double[]>();
    yList = new ArrayList<double[]>();
    roList = new ArrayList<Double>();
    double[] nextS;
    double[] nextY;

    //    double initGradNorm = ArrayMath.norm(grad);
    sayln("Iter: n <chooseDir> [(derivInDir) chooseNewPoint] newValue (relAvgImprovement)\n");

    boolean have_max = (maxIterations > 0);
    for (k = 0; ; k++) {

      double newValue = 0;
      try {
        say("Iter: " + k + " ");

        // compute search direction (dir)
        say("<");
        try {
          computeDir(dir, grad);
        } catch (SQNMinimizer.SurpriseConvergence s) {
          clearStuff();
          return x;
        }
        say("> ");

        if (M > 0 && sList.size() == M || sList.size() == 20) {
          nextS = sList.remove(0);
          nextY = yList.remove(0);
          roList.remove(0);
        } else {
          nextS = new double[x.length];
          nextY = new double[x.length];
        }

        //Update Gain
        gain = gainSchedule(k,tau)* initGain;

        //take the step
        //plusAndConstMult(x, dir, gain, newX);


        for(int i=0;i<x.length;i++){
          newX[i] = x[i] + gain*dir[i];
        }


       /*
              //Update the weights
        for(int i = 0; i < x.length; i++){
          double meta = 1-mu*dir[i]*v[i];

          if(0.5 > meta){
            gainVect[i] = gainVect[i]*0.5;
          }else{
            gainVect[i] = gainVect[i]*meta;
          }
          gainVect[i] = initGain;
        // NOW BEGINING iteration i+1
        //----------------------------
        //Update gain history
        v[i] = lam*(1+gainVect[i])*v[i] - gainVect[i]*(dir[i] + lam*Hdir[i]);
        //Get the next X
        newX[i] = x[i] + gainVect[i]*dir[i];
      }
       */

        // DONT perform line search
        //say("[");
        //newValue = lineSearch(dfunction, dir, x, newX, grad, value);
        //say("] ");
      } catch (OutOfMemoryError e) {
        sayln(" --- Reached memory limit.  Setting m and redoing iteration...");
        M = sList.size();
        k--;
        continue;
      }

      dfunction.recalculatePrevBatch = true;
      System.arraycopy(dfunction.derivativeAt(newX,dir,batchSize), 0, newGradSameBatch, 0, newGradSameBatch.length);

      dfunction.returnPreviousValues = true;
      System.arraycopy(dfunction.HdotVAt(newX,dir,batchSize), 0, Hdir, 0, Hdir.length);
      
      dfunction.hasNewVals = true;
      System.arraycopy(dfunction.derivativeAt(newX,batchSize), 0, newGrad, 0, newGrad.length);

      System.err.println("normdiff -" + ArrayMath.norm( ArrayMath.pairwiseSubtract(Hdir,grad) ));
      //dfunction.returnPreviousValues = true;  //Same deal Don't even try and compare x and v we know they're the same
      //System.arraycopy(dfunction.HdotVAt(newX,v,newGrad,batchSize) ,0,Hdir,0,Hdir.length);

      //say(nf.format(dfunction.valueAt(newX)));

      // compute s_k, y_k
      ro = 0;
      for(int i=0;i<x.length;i++){
        nextS[i] = newX[i] - x[i];
        //plusAndConstMult(newX, x, -1, nextS);
        nextY[i] = newGradSameBatch[i] - grad[i] + lambda*nextS[i];
        //plusAndConstMult(newGradSameBatch, grad, -1, nextY);
        ro += nextS[i]*nextY[i];
      }

      ro = 1.0 / ro;

      sList.add(nextS);
      yList.add(nextY);
      roList.add(ro);

      previousVals.add(ArrayMath.norm(newGrad));

      int size = previousVals.size();
      double previousVal = size == 10 ? previousVals.remove() : previousVals.peek();
      double averageNorm = (previousVal - newValue) / size;

      say(" (" + nf.format(averageNorm) + ")");


      //GOING TO NEED TO USE GRADIENT NORM FOR CONVERGENCE NOT THIS..
      if ((size > 5 && averageNorm < functionTolerance) ||
					(have_max && k >= maxIterations)) {
        System.err.println(" DONT TRUST RESULTS!!! ASK ALEX ABOUT CONVERGENCE CRITERA!");

        //Make sure that these critera don't get used.
        //System.exit(1);
        /*
        if(outputIterationsToFile){
          double finalVal = dfunction.valueAt(x);
          infoFile.println(" iterationCount ; " + k);
          infoFile.println(" finalValue ; " + finalVal);
          infoFile.println(finalVal);
          infoFile.close();
          file.close();
          System.err.println("Output Files Closed");

        }

        clearStuff();
        return newX;
        */
      }



      if(outputIterationsToFile && (k%outputFrequency == 0)) {
        calcTime = (System.currentTimeMillis()- calcTime);
        double curVal = dfunction.valueAt(x);
        say("TrueVal = " + nf.format(curVal));
        //say(" TrueValue{ " + curVal + " } ");
        file.println(k + " , " + curVal + " , " + averageNorm + " , " + calcTime + " , " + ArrayMath.norm(ArrayMath.pairwiseSubtract(x,newX)));
        calcTime = System.currentTimeMillis();
      }

      sayln("");

      if (monitor != null) {
        monitor.valueAt(newX);
      }

      // shift
      double[] temp = x;
      x = newX;
      newX = temp;

      System.arraycopy(newGrad, 0, grad, 0, newGrad.length);
      if (quiet) {
        System.err.print(".");
      }
    }
  }




    private void outputInitializeFiles(){

    if(outputIterationsToFile){
      try{
        file = new PrintWriter(new FileOutputStream("StochasticQN_b" + batchSize + "_g" + (initGain* 1000) + ".output"),true);
        infoFile = new PrintWriter(new FileOutputStream("StochasticQN_b" + batchSize + "_g" + (initGain* 1000) + ".info"),true);
      }
      catch (IOException e){
        System.err.println("Caught IOException outputing SMD data to file: " + e.getMessage());
        System.exit(1);
      }
    };
  }





 

  private void clearStuff() {
    sList = null;
    yList = null;
    roList = null;
  }

  private static class SurpriseConvergence extends Throwable {
    public SurpriseConvergence(String s) {
      super(s);
    }
  }

  private void sayln(String s) {
    if (!quiet) {
      System.err.println(s);
    }
  }

  private void say(String s) {
    if (!quiet) {
      System.err.print(s);
    }
  }



}
